'
' -----[ Program Description ]--------------------------------------------------
'
' This program implements an PID algorithm in BASCOM
'
' This File is partialy based on a program from ?????? and
' I do not intent to touch any of his rights !!!
' So for correct use of his copyrights please inform as marked below!
' ?????????
'
'
' -----[ Disclaimer ]-----------------------------------------------------------
'
' This example is offered on an "AS IS" basis, no warranty expressed or implied.
' The programers disclaim liability of any damages associated with the use of
' the hardware or software described herein. You use it on your own risk.
' I'm not able to provide any free support.
'
' Copyright (c) 2001 Mike Eitel all rights reserved
'
' -----[ Revision History ]-----------------------------------------------------
'
' 010429 - Version AVRPID Ver .9    Basic PID functionality   Mike Eitel
'
' -----[ Aliases ]--------------------------------------------------------------
'$sim                                              ' Helps testing in a simulation

' -----[ Constant ]-------------------------------------------------------------
   Const Regulators = 7                           ' Nr of regulators
                                                  ' max seven in this example
   'Const Kp = .85                                 ' Proportional factor
   'Const Ki = .67                                 ' Integration factor
   'Const Kd = .10                                 ' Derivation factor
' -----[ Variables ]------------------------------------------------------------
   Dim Auto_mode As Byte                          ' Switch pointer regulators on ?
   Dim Manual_value(regulators) As Single         ' Output if not regulating

   Dim Sp(regulators) As Single                      ' Setpoint
   Dim Pv(regulators) As Single                   ' Process Value
   Dim Cv(regulators) As Single                   ' PID output

   Dim First_execution(regulators) As Byte        ' First start recognition
   Dim Initial_error(regulators) As Single        ' Startup difference


   Dim Reg_nr As Byte                             ' Used regulator

   Dim A As Byte                                  ' Tmp for random
   Dim B As Single                                'Tmp for random


' -----[ Start of program ]-----------------------------------------------------
' -----[ Start of program ]-----------------------------------------------------
' -----[ Start of program ]-----------------------------------------------------
   For Reg_nr = 1 To Regulators
      Auto_mode.reg_nr = 1                        ' Permanent running choosen
                                                  ' this .byte alows max 7 PID's !!
      First_execution(reg_nr) = 0                 ' Set permanent running algorithm
      Manual_value(reg_nr) = 40                   ' Output if not regulating = on
      Sp(reg_nr) = 37 + Reg_nr                    ' Value to aim to
   Next Regulators

   Gosub Parameters                               ' Fill the parameters

Cyclic:
' -----[ Start of endless running program ]-------------------------------------
   If Reg_nr > Regulators Then                    ' Give all regulators a chance
      Reg_nr = 1                                  ' in a simple loop
   End If

   Gosub Regulator                                ' Call the PID allgorithm
   Gosub Object                                   ' Call the simulated outer loop
Goto Cyclic
   ' -----[ End of endless running program ]------------------------------------
   ' -----[ End of endless running program ]------------------------------------
   ' -----[ End of endless running program ]------------------------------------


   ' ---------------------------------------------------------------------------
   ' -----------------------------[ Subroutines ]-------------------------------
   ' ---------------------------------------------------------------------------
Object:
   ' -----[ Start of simulated regulation loop ]--------------------------------
    Pv(reg_nr) = Pv(reg_nr) + Cv(reg_nr)          ' linear function used

    If Pv(reg_nr) = Sp(reg_nr) Then               ' When PV=SP then make a
      A = Rnd(100)                                ' random SP jump
      Sp(reg_nr) = 1 * A
    End If

    Reg_nr = Reg_nr + 1                           ' Next regulator
Return

Parameters:
   ' -----[ Start of PID parameters]--------------------------------------------
   ' -----[ Variables ]---------------------------------------------------------
   Dim Kp(regulators) As Single                   ' Proportional factor
   Dim Ki(regulators) As Single                   ' Integration factor
   Dim Kd(regulators) As Single                   ' Derrivation factor
   ' -----[ Code ]--------------------------------------------------------------
   For Reg_nr = 1 To Regulators
      A = Rnd(100)
      B = A * 0.01
      Kp(reg_nr) = .85                            ' Proportional factor
      Ki(reg_nr) = .67 + B                        ' Integration factor
      Kd(reg_nr) = .10                            ' Derivation factor
   Next Regulators
Return

Regulator:
   ' -----[ Start of PID Regulator]---------------------------------------------
   ' If only one regulator us used these variables should be declared local
   ' -----[ Constant ]----------------------------------------------------------
   'Const Kp = .85                              ' Proportional factor
   'Const Ki = .67                              ' Integration factor
   'Const Kd = .10                              ' Derivation factor
   ' -----[ Variables ]---------------------------------------------------------
   'Dim Sp(regulators) As Single                ' Setpoint
   'Dim Pv(regulators) As Single                ' Process Value
   'Dim Cv(regulators) As Single                ' PID output
   '
   'Dim First_execution(regulators) As Byte     ' First start recognition
   'Dim Initial_error(regulators) As Single     ' Startup difference
   Dim Last_pv(regulators) As Single           ' Last PV
   Dim Last_sp(regulators) As Single           ' Last SP
   Dim Sum_error(regulators) As Single         ' Summed error value
   Dim D_pv(regulators) As Single              ' Derrivated delta PV

   Dim Error As Single                         ' Difference between SP and PV
   Dim Pterm As Single                         ' Proportional calculated part
   Dim Iterm As Single                         ' Integrated calculated part
   Dim Dterm As Single                         ' Derivated calculated part
   ' -----[ Code ]--------------------------------------------------------------

   If Auto_mode.reg_nr = 1 Then
      ' -------- Regulating modus
      Error = Sp(reg_nr) - Pv(reg_nr)
      Sum_error(reg_nr) = Sum_error(reg_nr) + Error
      Iterm = Ki(reg_nr) * Sum_error(reg_nr)      ' Integrated CV part

      ' -------- First time startup
      If First_execution(reg_nr) < 2 Then
         If First_execution(reg_nr) = 0 Then
            Sum_error(reg_nr) = Manual_value(reg_nr) / Ki(reg_nr)
            First_execution(reg_nr) = 1
            Initial_error(reg_nr) = Error
         End If
         Pterm = 0
         Dterm = 0
         If Initial_error(reg_nr) > 0 And Error < 0 Then
            First_execution(reg_nr) = 2
            Last_pv(reg_nr) = Pv(reg_nr)
         End If
         If Initial_error(reg_nr) < 0 And Error > 0 Then
            First_execution(reg_nr) = 2
            Last_pv(reg_nr) = Pv(reg_nr)
         End If
         Last_sp(reg_nr) = Sp(reg_nr)

      ' -------- Normal calculation loop
      Else
         D_pv(reg_nr) = Last_pv(reg_nr) - Pv(reg_nr)
         Last_pv(reg_nr) = Pv(reg_nr)
         Dterm = Kd(reg_nr) * D_pv(reg_nr)        ' Derivated CV part
         If Sp(reg_nr) = Last_sp(reg_nr) Then
            ' -------- Normal loop when setpoint not changed
            Pterm = Kp(reg_nr) * Error            ' Proportional CV part
            ' -------- Loop when setpoint changed
            Else
            Pterm = 0
            Dterm = 0
            If Sp(reg_nr) > Last_sp(reg_nr) And Pv(reg_nr) > Sp(reg_nr) Then
               Last_sp(reg_nr) = Sp(reg_nr)
               Last_pv(reg_nr) = Pv(reg_nr)
            End If
            If Sp(reg_nr) < Last_sp(reg_nr) And Pv(reg_nr) < Sp(reg_nr) Then
               Last_sp(reg_nr) = Sp(reg_nr)
               Last_pv(reg_nr) = Pv(reg_nr)
            End If
         End If                                   ' Enf of SP change seperation                                   '
      End If                                      ' Enf of first time running seperation                                  '

      Cv(reg_nr) = Pterm + Iterm                  ' Summing of the tree
      Cv(reg_nr) = Cv(reg_nr) + Dterm             ' calculated terms

   ' -------- Forced modus
   Else                                           ' When running in non regulationg modus
      Cv(reg_nr) = Manual_value(reg_nr)           ' Set output to predefined value
      First_execution(reg_nr) = 0                 ' restart bumpless
   End If
Return
